package com.dotsource.intershop.sofort.payment;

import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.util.Arrays;
import java.util.Date;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;

import org.apache.commons.lang.StringUtils;

import static com.dotsource.intershop.sofort.payment.SofortConstants.*;

import com.intershop.beehive.bts.capi.payment.PaymentAuthorizationException;
import com.intershop.beehive.bts.capi.payment.PaymentHistory;
import com.intershop.beehive.bts.capi.payment.PaymentInstrumentInfo;
import com.intershop.beehive.bts.capi.payment.PaymentServiceCapabilities;
import com.intershop.beehive.bts.capi.payment.PaymentTransaction;
import com.intershop.beehive.bts.capi.payment.PaymentTransactionConstants;
import com.intershop.beehive.configuration.capi.common.Configuration;
import com.intershop.beehive.core.capi.domain.Domain;
import com.intershop.beehive.core.capi.environment.PropertyMgr;
import com.intershop.beehive.core.capi.log.Logger;
import com.intershop.beehive.core.capi.request.Request;
import com.intershop.beehive.core.capi.request.RequestInformation;
import com.intershop.beehive.core.capi.url.Action;
import com.intershop.beehive.core.capi.url.Parameters;
import com.intershop.beehive.core.capi.url.URLComposition;
import com.intershop.component.payment.capi.AbstractPaymentService;
import com.intershop.component.payment.capi.PaymentServiceProxy;
import com.intershop.component.service.capi.service.ServiceConfigurationBO;
import com.sofort.lib.core.products.request.parts.Notification;
import com.sofort.lib.core.products.response.parts.FailureMessage;
import com.sofort.lib.payment.DefaultSofortLibPayment;
import com.sofort.lib.payment.SofortLibPayment;
import com.sofort.lib.payment.products.request.PaymentRequest;
import com.sofort.lib.payment.products.request.PaymentTransactionDetailsRequest;
import com.sofort.lib.payment.products.request.SofortPayment;
import com.sofort.lib.payment.products.response.PaymentResponse;
import com.sofort.lib.payment.products.response.PaymentTransactionDetailsResponse;
import com.sofort.lib.payment.products.response.parts.PaymentTransactionDetails;

public class SofortPaymentService extends AbstractPaymentService
{
    // port used for https requests
    private static final String SECURE_PORT = PropertyMgr.getInstance().getProperty("mgmt.cluster.webserver.https");
    
    public SofortPaymentService(ServiceConfigurationBO serviceConfiguration, Class<?> serviceInterface)
    {
        super(serviceConfiguration, serviceInterface);
    }

    @Override
    public PaymentServiceCapabilities getCapabilities()
    {
        return new SofortPaymentServiceCapabilities(new PaymentServiceProxy(this));
    }

    @Override
    public String getID()
    {
        return SOFORT_ID;
    }

    public Map<String, Object> prepareEditCheckout(PaymentInstrumentInfo paymentInstrumentInfo, Domain domain,
                    String targetPipeline)
    {
        Map<String, Object> result = new HashMap<String, Object>();
        result.put("Display_CheckoutIncludeMiniTemplate", "false");

        return result;
    }

    protected SofortLibPayment getSofortLibPaymentObject()
    {
        String tmpProjectID = getCredentialFromConfiguration(PREF_CONFIG_KEY_IDX_PROJECT_ID);
        String tmpProjectPW = getCredentialFromConfiguration(PREF_CONFIG_KEY_IDX_API_KEY);
        String tmpUserID = getCredentialFromConfiguration(PREF_CONFIG_KEY_IDX_USER_ID);

        if (null == tmpProjectID || null == tmpProjectPW || null == tmpUserID)
        {
            Logger.error(this,"Could not get credentials from "+SOFORT_ID+" payment configuration key!");
            return null;
        }

        // initial setup of SofortLibPament
        SofortLibPayment sofortLibPayment = new DefaultSofortLibPayment(Integer.parseInt(tmpUserID), tmpProjectPW);

        return sofortLibPayment;
    }

    /**
     * Intershop Payment Framework Method to get the redirect URL
     * 
     * extended by parameters for the certain service
     */
    public String calculateRedirectURLForAuthorization(PaymentTransaction transaction, Request currentRequest,
                    String mailAddress, String targetPipeline) throws PaymentAuthorizationException
    {
        Map<String, String> historyParameter = new HashMap<String, String>();
        String tmpProjectID = getCredentialFromConfiguration(PREF_CONFIG_KEY_IDX_PROJECT_ID);

        final SofortLibPayment sofortLibPayment = getSofortLibPaymentObject();

        // get currency mnemonic value from transaction for sofort-request
        String currency = transaction.getAmount().getCurrencyMnemonic();

        // get amount value from transaction for sofort-request
        Double amount = transaction.getAmount().getValue().doubleValue();
        
        // get reason strings
        String reason1String = this.getConfiguration().getString(PREF_REASON_1);
        String reason2String = this.getConfiguration().getString(PREF_REASON_2);

        // setting the buyer/consumer protection in the payment request is not mandatory anymore 
        SofortPayment sofortPayment = new SofortPayment();
        
        // authorization request for sofort
        PaymentRequest paymentRequest = new PaymentRequest(Integer.parseInt(tmpProjectID), amount, currency,
                        Arrays.asList(convertReasonString(reason1String, transaction),
                                      convertReasonString(reason2String, transaction)), 
                        sofortPayment);
        
        /*
         * 1st step -> start sofort payment, check for errors and warnings and
         * use the received payment url for redirection of the buyer
         */

        Parameters urlparams = new Parameters();

        urlparams.addParameter("TransactionUUID", transaction.getUUID());
        final String notifyUrl = createReturnUrl(currentRequest, currentRequest.getExecutionSite().getDomainName(),
                        "ViewSofortPayment-CallbackNotify", urlparams);

        urlparams.addParameter("JumpTarget", targetPipeline);
        final String returnUrl = createReturnUrl(currentRequest, currentRequest.getExecutionSite().getDomainName(),
                        "ViewSofortPayment-CallbackRedirect", urlparams);
        final String abortUrl = createReturnUrl(currentRequest, currentRequest.getExecutionSite().getDomainName(),
                        "ViewSofortPayment-CallbackAbort", urlparams);

        paymentRequest.setNotificationUrls(Arrays.asList(new Notification(notifyUrl)));
        paymentRequest.setAbortUrl(abortUrl);
        paymentRequest.setSuccessUrl(returnUrl);
        
        // authorization call to sofort payment provider
        Logger.debug(this,"Sending initial payment request.\n\tNotificationURL = "+notifyUrl+"\n\tAbortURL = "+abortUrl+"\n\tReturnUrl = "+returnUrl);
        PaymentResponse paymentResponse = sofortLibPayment.sendPaymentRequest(paymentRequest);

        if (paymentResponse.hasResponseErrors() || paymentResponse.hasResponseWarnings())
        {
            // check and handle the response errors and warnings
            List<FailureMessage> errors = paymentResponse.getResponseErrors();
            if (errors != null && !errors.isEmpty())
            {
                for (FailureMessage message : errors)
                {
                    Logger.error(this,
                                    "Response Error Code '" + message.getCode() + "' in Field '" + message.getField()
                                                    + "': " + message.getMessage());
                }
            }

            List<FailureMessage> warnings = paymentResponse.getResponseWarnings();
            if (warnings != null && !warnings.isEmpty())
            {
                for (FailureMessage message : warnings)
                {
                    Logger.warn(this,
                                    "Response Warning Code '" + message.getCode() + "' in Field '" + message.getField()
                                                    + "': " + message.getMessage());
                }
            }
        }

        if (paymentResponse.hasNewPaymentWarnings())
        {
            // check the new payment warnings
            List<FailureMessage> warnings = paymentResponse.getNewPaymentWarnings();
            if (warnings != null && !warnings.isEmpty())
            {
                for (FailureMessage message : warnings)
                {
                    Logger.warn(this,
                                    "Payment Warning Code '" + message.getCode() + "' in Field '" + message.getField()
                                                    + "': " + message.getMessage());
                }
            }
        }

        String txid = paymentResponse.getTransId();
        transaction.setServiceTransactionID(txid);

        historyParameter.put("history.payment.sofort.status", "history.payment.sofort.status.temporary.1648");
        historyParameter.put("history.payment.sofort.message", "history.payment.sofort.text.temporary.1801");
        historyParameter.put("history.payment.sofort.transid", txid);
        historyParameter.put("history.payment.sofort.url", paymentResponse.getPaymentUrl());
        
        Logger.debug(this,"Got initial payment response.\n\tTransactionID="+txid+"\n\tPaymentURL="+paymentResponse.getPaymentUrl());
        
        createPaymentHistoryEvent(transaction, SOFO_MAP_STATE_TEMPORARY, historyParameter);
        transaction.putString(TRANSACTION_SOFO_AV_CURRENT_MAPPED_STATE, SHOP_MAPPED_STATE_TEMPORARY);
        return paymentResponse.getPaymentUrl();
    }

    /**
     * creates a return URL for the PaymentService Redirect
     */
    private static String createReturnUrl(Request currentRequest, String siteName, String returnPipeline,
                    Parameters parameters)
    {
        String currency = currentRequest.getCurrencyCode();
        String protocol = "https";
        String hostname = (String)currentRequest.getObject(RequestInformation.HOST);
        String localeId = (String)currentRequest.getObject(RequestInformation.REQUESTED_LOCALE);
        String servergroup = (String)currentRequest.getObject(RequestInformation.SERVER_GROUP);

        Action returnAction = new Action(returnPipeline, servergroup, siteName, localeId, currency);

        return new URLComposition().createSessionlessURL(protocol, hostname, SECURE_PORT, returnAction, parameters);
    }

    
    /**
     * 
     * calls a request to Sofortueberweisung and returns the request result
     */
    public SofortPaymentStatus getStatusResult(PaymentTransaction paymentTransaction, String requestType)
    {
        final SofortLibPayment sofortLibPayment = getSofortLibPaymentObject();

        PaymentTransactionDetailsRequest transactionRequest = new PaymentTransactionDetailsRequest().setTransIds(Arrays
                        .asList(paymentTransaction.getServiceTransactionID()));

        if (transactionRequest != null)
        {
            Logger.debug(this,"Sending status request triggered by "+requestType+".\tTransactionID="+paymentTransaction.getServiceTransactionID());
            PaymentTransactionDetailsResponse transactionDetailsResponse = sofortLibPayment
                            .sendTransactionDetailsRequest(transactionRequest);
            PaymentTransactionDetails detailsPayment = transactionDetailsResponse.getTransactions().get(0);

            Logger.debug(this,"Got status response.\tTransactionID="+detailsPayment.getTransId()+"\tStatus="+detailsPayment.getStatus().name()+
                            "\tReason="+detailsPayment.getStatusReason().name()+"\tCurrency="+detailsPayment.getCurrencyCode()+
                            "\tAmount="+detailsPayment.getAmount()+"\tAmountRefunded="+detailsPayment.getAmountRefunded());

            if (detailsPayment.getStatus().name().equals(SOFO_TRANSACTION_STATE_LOSS)
                            || detailsPayment.getStatus().name().equals(SOFO_TRANSACTION_STATE_PENDING)
                            || detailsPayment.getStatus().name().equals(SOFO_TRANSACTION_STATE_RECEIVED)
                            || detailsPayment.getStatus().name().equals(SOFO_TRANSACTION_STATE_REFUNDED)
                            || detailsPayment.getStatus().name().equals(SOFO_TRANSACTION_STATE_UNTRACEABLE))
            {
                return new SofortPaymentStatus(detailsPayment.getStatus().name(), detailsPayment.getStatusReason().name(), detailsPayment.getTransId(),
                                detailsPayment.getCurrencyCode(), detailsPayment.getAmount(), detailsPayment.getAmountRefunded(), requestType);
            }
            
        } else {
            Logger.error(this,"Error creating status request.\nTransactionID="+paymentTransaction.getServiceTransactionID()+"\nRequestType="+requestType);
        }
        
        return new SofortPaymentStatus(STATUS_RESULT_FAILED, "unknown", "unknown", "unknown", 0.0, 0.0, requestType);
    }

    
    /**
     * 
     * process the mapped action defined for abort if the CallbackAbort pipeline is called
     */
    public String abortPayment(PaymentTransaction transaction, String requestType)
    {
        Configuration serviceconf = this.getConfiguration();
        Map<String, String> historyParameter = new HashMap<String, String>();
        String mappedState = null;
        
        historyParameter.put("history.payment.sofort.status", "history.payment.sofort.status.aborted.1650");
        historyParameter.put("history.payment.sofort.message", "history.payment.sofort.text.aborted.1804");

        Integer refundAmountValue;
        
        try {
            refundAmountValue = new Integer(transaction.getString(TRANSACTION_SOFO_AV_REFUNDED_TOTAL_VALUE));
        } catch(NumberFormatException e) {
            refundAmountValue = 0;
        }
                
        SofortPaymentStatus status = new SofortPaymentStatus(null, null, transaction.getServiceTransactionID(), 
            transaction.getAmount().getCurrencyMnemonic(), transaction.getAmount().getValue().intValue(), refundAmountValue, requestType);        
        
        mappedState = serviceconf.getString(SOFO_MAP_STATE_ABORTED);
        
        if(processStatusSuccessfully(transaction, status, requestType, historyParameter, SOFO_MAP_STATE_ABORTED, mappedState))
        {
            return mappedState;        
        } else {
            return SHOP_MAPPED_STATE_FAILED;
        }        
    }


    /**
     * 
     * status change logic
     */
    public String updateStatus(PaymentTransaction transaction, SofortPaymentStatus status, String requestType)
    {
        Configuration serviceconf = this.getConfiguration();
        
        Map<String, String> historyParameter = new HashMap<String, String>();
        String inputState = null;
        String mappedState = null;
        
        switch(status.getReason())
        {
            case SOFO_TRANSACTION_STATE_REASON_NOT_CREDITED_YET: 
                historyParameter.put("history.payment.sofort.status", "history.payment.sofort.status.confirmed.1649");
                historyParameter.put("history.payment.sofort.message", "history.payment.sofort.text.confirmed.1802");
                historyParameter.put("history.payment.sofort.transid", transaction.getServiceTransactionID());
                inputState = SOFO_MAP_STATE_CONFIRMED;
                mappedState = serviceconf.getString(SOFO_MAP_STATE_CONFIRMED); 
                break;
            case SOFO_TRANSACTION_STATE_REASON_SBAN:
                historyParameter.put("history.payment.sofort.status", "history.payment.sofort.status.confirmed.1649");
                historyParameter.put("history.payment.sofort.message", "history.payment.sofort.text.confirmed.1802");                
                historyParameter.put("history.payment.sofort.transid", transaction.getServiceTransactionID());
                inputState = SOFO_MAP_STATE_CONFIRMED;
                mappedState = serviceconf.getString(SOFO_MAP_STATE_CONFIRMED); 
                break;
            case SOFO_TRANSACTION_STATE_REASON_NOT_CREDITED: 
                historyParameter.put("history.payment.sofort.status", "history.payment.sofort.status.checkorder.1651");
                historyParameter.put("history.payment.sofort.message", "history.payment.sofort.text.checkorder.1805");
                inputState = SOFO_MAP_STATE_CHECK_ORDER;
                mappedState = serviceconf.getString(SOFO_MAP_STATE_CHECK_ORDER); 
                break;
            case SOFO_TRANSACTION_STATE_REASON_CREDITED: 
                historyParameter.put("history.payment.sofort.status", "history.payment.sofort.status.received");
                historyParameter.put("history.payment.sofort.message", "history.payment.sofort.text.received.1803");
                inputState = SOFO_MAP_STATE_RECEIVED;
                mappedState = serviceconf.getString(SOFO_MAP_STATE_RECEIVED); 
                break;
            case SOFO_TRANSACTION_STATE_REASON_COMPENSATION: 
                historyParameter.put("history.payment.sofort.status", "history.payment.sofort.status.refund.1652");
                historyParameter.put("history.payment.sofort.message", "history.payment.sofort.text.partialrefund.1807");
                historyParameter.put("history.payment.sofort.partialamount", Double.toString(status.getAmountRefunded()));
                inputState = SOFO_MAP_STATE_PARTIAL_REFUNDED;
                mappedState = serviceconf.getString(SOFO_MAP_STATE_PARTIAL_REFUNDED); 
                break;
            case SOFO_TRANSACTION_STATE_REASON_REFUNDED: 
                historyParameter.put("history.payment.sofort.status", "history.payment.sofort.status.refund.1652");
                historyParameter.put("history.payment.sofort.message", "history.payment.sofort.text.refund.1806");
                inputState = SOFO_MAP_STATE_REFUNDED;
                mappedState = serviceconf.getString(SOFO_MAP_STATE_REFUNDED); 
                break;
            default:
                Logger.debug(this,"(UpdateStatus-"+requestType+") Unhandled status reason: "+status.getReason()+" in Transaction "+transaction.getServiceTransactionID());
                break;
        }
        
        if(processStatusSuccessfully(transaction, status, requestType, historyParameter, inputState, mappedState))
        {
            return mappedState;        
        } else {
            return SHOP_MAPPED_STATE_FAILED;
        }
        
    }

    /**
     * 
     * process the mapped shop state
     */
    private Boolean processStatusSuccessfully(PaymentTransaction transaction, SofortPaymentStatus status, String requestType,
                    Map<String, String> historyParameter, String inputState, String mappedState)
    {
        Boolean writeHistory = true;
        
        if (transaction != null && !StringUtils.isEmpty(mappedState))
        {
            
            if(SofortPaymentStateModel.isChangeAllowed(transaction.getString(TRANSACTION_SOFO_AV_CURRENT_MAPPED_STATE), mappedState))
            {                
                Logger.debug(this,"(UpdateStatus-"+requestType+") "+transaction.getString(TRANSACTION_SOFO_AV_CURRENT_MAPPED_STATE)
                                +" --> "+mappedState+" in Transaction "+transaction.getServiceTransactionID());
                transaction.putString(TRANSACTION_SOFO_AV_CURRENT_MAPPED_STATE, mappedState);
                
                switch(mappedState)
                {
                    case SHOP_MAPPED_STATE_TEMPORARY :
                        transaction.setStatus(PaymentTransactionConstants.PT_CREATED);
                        break;
                    case SHOP_MAPPED_STATE_CONFIRMED :
                        transaction.setStatus(PaymentTransactionConstants.PT_CAPTURED);
                        break;
                    case SHOP_MAPPED_STATE_ABORTED :
                        // there is no cancelled status for transactions - put it back to the initial state
                        transaction.setStatus(PaymentTransactionConstants.PT_CREATED);
                        break;
                    case SHOP_MAPPED_STATE_CHECK_ORDER :
                        transaction.setStatus(PaymentTransactionConstants.PT_CAPTUREVOIDED);
                        break;
                    case SHOP_MAPPED_STATE_REFUNDED :
                        // refund value will be updated
                        transaction.putDouble(TRANSACTION_SOFO_AV_REFUNDED_TOTAL_VALUE, status.getAmountRefunded());
                        transaction.setStatus(PaymentTransactionConstants.PT_REFUNDED);
                        break;
                    case SHOP_MAPPED_STATE_PARTIAL_REFUNDED :
                        // status change to refund on partial refund / refund value will be updated
                        transaction.putDouble(TRANSACTION_SOFO_AV_REFUNDED_TOTAL_VALUE, status.getAmountRefunded());
                        transaction.setStatus(PaymentTransactionConstants.PT_REFUNDED);
                        break;
                    case SHOP_MAPPED_STATE_NO_CHANGE :
                        // no action required
                        break;
                    default:
                        writeHistory = false;
                        break;
                }
                                        
            } else {
                writeHistory = false;
                Logger.debug(this,"(UpdateStatus-"+requestType+") Unsupported status change for "+transaction.getString(TRANSACTION_SOFO_AV_CURRENT_MAPPED_STATE)
                                +" --> "+mappedState+" in Transaction "+transaction.getServiceTransactionID());
            }

            if(writeHistory)
            {
                createPaymentHistoryEvent(transaction, inputState, historyParameter);
                return true;
            }
            
        } else {
            if(transaction == null) {
                Logger.error(this,"(UpdateStatus) Transaction is null!");
            } else {
                Logger.error(this,"(UpdateStatus) Mapped status is null!");
            }
        }        
        return false;
    }
    
    
    /**
     * 
     * placeholder logic for strings in the reason fields of the initial payment request;
     * also formats the string depending on various restrictions needed for acceptance on SOFO side
     */
    public String convertReasonString(String inputReason, PaymentTransaction transaction)
    {
        String outputReason;
        if(StringUtils.isEmpty(inputReason))
        {
            outputReason =  StringUtils.EMPTY;
        } else {
            outputReason = inputReason;
        }
        
        // fill placeholders
        outputReason = outputReason.replaceAll(PH_START+PH_ORDER_ID+PH_END, transaction.getOrder().getDocumentNo());
        outputReason = outputReason.replaceAll(PH_START+PH_ORDER_DATE+PH_END, getFormattedDate(transaction.getOrder().getCreationDate(),"yyMMddHHmmssZ"));
        outputReason = outputReason.replaceAll(PH_START+PH_CUSTOMER_ID+PH_END, transaction.getOrder().getBuyerNo());
        outputReason = outputReason.replaceAll(PH_START+PH_CUSTOMER_NAME+PH_END, 
                        transaction.getOrder().getBuyerFirstName()+" "+transaction.getOrder().getBuyerLastName());
        outputReason = outputReason.replaceAll(PH_START+PH_CUSTOMER_COMPANY+PH_END, transaction.getOrder().getInvoiceToAddress().getCompanyName());
        outputReason = outputReason.replaceAll(PH_START+PH_TRANSACTION_ID+PH_END, "-TRANSACTION-");  // will be replaced by the service
        
        // char restrictions by API as written in documentation (only a-z, A-Z, 0-9, sz and umlauts - other chars will be skipped)
        outputReason = outputReason.replaceAll("[^a-zA-Z0-9\\,\\+\\-\\.\\ \\u00df\\u00c4\\u00e4\\u00d6\\u00f6\\u00dc\\u00fc]", StringUtils.EMPTY);
        
        // check maximum length
        if(outputReason.length() > MAX_REASON_STRING_LENGTH)
        {
            Logger.warn(this,"Reason string of order "+transaction.getOrder().getDocumentNo()+" exceed maximal length of "+MAX_REASON_STRING_LENGTH+" - truncated!\n"
                            +"Complete content: "+outputReason);
            outputReason = outputReason.substring(0, MAX_REASON_STRING_LENGTH);
        }
        // check minimum length
        if(outputReason.length() == 0)
        {
            Logger.warn(this,"Reason string of order "+transaction.getOrder().getDocumentNo()+" is empty - set to 'undefined'!");
            outputReason = "undefined";
        }
        return outputReason;
    }
    
    
    /**
     * 
     * returns a formatted date string
     */
    public String getFormattedDate(Date inputDate, String format)
    {
        final DateFormat df = new SimpleDateFormat(format);
        final String formattedDateString = df.format(inputDate);        
        return formattedDateString;
    }

    
    /**
     * 
     * adds a new entry to the payment history
     */
    public void createPaymentHistoryEvent(PaymentTransaction transaction, String event, Map<String, String> parameters)
    {
        PaymentHistory ph = transaction.addToPaymentHistory(event);

        for (Entry<String, String> entry : parameters.entrySet())
        {
            ph.putString(entry.getKey(), entry.getValue());
        }
    }
        
    
    /**
     * method is mandatory for Payment History Integration
     */
    @Override
    public Map<String, Object> prepareBOHistoryView(PaymentTransaction paymenttransaction, Domain domain, String s)
    {
        HashMap<String, Object> historyView = new HashMap<String, Object>();
        return historyView;
    }
    
    
    /**
     * 
     * extracts credential attributes from combined string
     */
    private String getCredentialFromConfiguration(int cfgIndex)
    {
        String cfgKey = this.getConfiguration().getString(PREF_CONFIG_KEY);
        if(cfgKey == null || cfgKey.length() == 0)
        {
            return null;
        }
        
        String[] cfgArray = cfgKey.split(":");
        if(cfgArray.length != PREF_CONFIG_KEY_SIZE)
        {
            return null;
        }
        
        return cfgArray[cfgIndex];
    }
    
}
